package combinatorParsing2

/** A class of generic parsers
 *  This class defines:
 *   - a type Parser.
 *   - generic parser combinators: sequence, alternative, result production, 
 *     option, repetition, etc
 *   - primitive parser generators for success and failure.
 */
trait Parsers {

  /** the type of input elements */
  type Elem

  /** The parser input is an abstract reader of input elements */
  type Input = Reader[Elem]

  /** A base class for parser results. 
   *  Results are either successes or failures (see below).
   *  Successes carry a result, of type `T'.
   */
  class ParseResult[+T]

  /** The success case of ParseResult 
   *  @param out    The parser's output (which is a list of terminals or non-terminals)
   *  @param in     The parser's remaining input
   */
  case class Success[+T](result: T, in: Input) extends ParseResult[T] {
    /** The toString method of a Success */
    override def toString = "parsed: "+result
  }

  /** The failure case of ParseResult 
   *  @param msg    An error message string describing the failure
   *  @param in     The parser's unconsumed input at the point where the failure occured
   */
  case class Failure(msg: String, in: Input) extends ParseResult[Nothing] {
    var retry = true
    def noRetry = { retry = false; this }
    
    /** The toString method of a Failure yields an error message */
    override def toString = "["+in.pos+"] error: "+msg+"\n\n"+in.pos.longString 
  }

  /** The root class of parsers. 
   *  Parsers are functions from the Input type to ParseResult 
   */
  abstract class Parser[+T] extends (Input => ParseResult[T]) {

    /** An unspecified method the defines the parser function.
     */
    def apply(in: Input): ParseResult[T]

    /** A parser combinator for sequential composition 
     *
     *  `p ++ q' succeeds if `p' succeeds and `q' succeeds on the input
     *          left over by `p'.
     */
    def ++ [U](q: => Parser[U]): Parser[Pair[T, U]] = new Parser[Pair[T, U]] {
      def apply(in: Input) = seq(Parser.this, q)((x, y) => Pair(x, y))(true)(in)
    }

    /** A parser combinator for sequential composition with a unit-parser 
     *
     *  `p ++ q' succeeds if `p' succeeds and `q' succeeds on the input
     *          left over by `p'.
     */
    def ++ [Q <% UnitParser](q: Q): Parser[T] = new Parser[T] {
      def apply(in: Input) = seq(Parser.this, q)((x, y) => x)(true)(in)
    }
    
    def +++ [U](q: => Parser[U]): Parser[Pair[T, U]] = new Parser[Pair[T, U]] with OnceParser[Pair[T, U]] {
      def apply(in: Input) = seq(Parser.this, q)((x, y) => Pair(x, y))(false)(in)
    }

    def +++ [Q <% UnitParser](q: Q): Parser[T] = new Parser[T] with OnceParser[T] {
      def apply(in: Input) = seq(Parser.this, q)((x, y) => x)(false)(in)
    }

    /** A parser combinator for alternative composition 
     *
     *  `p | q' succeeds if `p' succeeds or `q' succeeds
     */
    def | [U >: T](q: => Parser[U]): Parser[U] = new Parser[U] {
      def apply(in: Input) = Parser.this(in) match {
        case s1 @ Success(_, _) => s1
        case f1 @ Failure(_, in1) =>
          if (f1.retry) q(in) match {
            case s2 @ Success(_, _) => s2
            case f2 @ Failure(_, in2) => if (in2.pos < in1.pos) f1 else f2
          } else
            f1
      }
    }

    /** A parser combinator for function application 
     *
     *  `p ^^ f' succeeds if `p' succeeds; it returns `f' applied to the result of `p'.
     */
    def ^^ [U](f: T => U): Parser[U] = new Parser[U] {
      def apply(in: Input) = mapfun(Parser.this)(f)(in)
    }     
  }

  /** The root class of special parsers returning the trivial result `unit'
   *  These compose differently from normal parsers in that the `unit'
   *  result in a sequential or function composition is dropped.
   */
  abstract class UnitParser extends (Input => ParseResult[unit]) {

    /** An unspecified method the defines the parser function.
     */
    def apply(in: Input): ParseResult[unit]

    /** A parser combinator for sequential composition 
     *
     *  `p ++ q' succeeds if `p' succeeds and `q' succeeds on the input
     *          left over by `p'.
     */
    def ++ [U](q: => Parser[U]): Parser[U] = new Parser[U] {
      def apply(in: Input): ParseResult[U] = seq(UnitParser.this, q)((x, y) => y)(true)(in)
    }
    
    /** A parser combinator for sequential composition with a unit-parser 
     *
     *  `p ++ q' succeeds if `p' succeeds and `q' succeeds on the input
     *          left over by `p'.
     */
    def ++ [A <% UnitParser](q: A): UnitParser = new UnitParser {
      def apply(in: Input): ParseResult[unit] = seq(UnitParser.this, q)((x, y) => y)(true)(in)
    }

    def +++ [U](q: => Parser[U]): Parser[U] = new Parser[U] with OnceParser[U] {
      def apply(in: Input) = seq(UnitParser.this, q)((x, y) => y)(false)(in)
    }

    def +++ [Q <% UnitParser](q: Q): UnitParser = new UnitParser with UnitOnceParser {
      def apply(in: Input) = seq(UnitParser.this, q)((x, y) => y)(false)(in)
    }

    /** A parser combinator for function application 
     *
     *  `p ^^ f' succeeds if `p' succeeds; it returns `f' applied to the result of `p'.
     */
    def ^^ [U](v: U): Parser[U] = new Parser[U] {
      def apply(in: Input) = mapfun(UnitParser.this)(x => v)(in)
    }
  }

  /** A helper method for sequential composition of (unit-)parsers */
  private def seq[T, U, V](p: Input => ParseResult[T], q: Input => ParseResult[U])
                          (compose: (T, U) => V)
                          (retry: Boolean)
                          (in: Input): ParseResult[V] = p(in) match {
    case Success(x, in1) => 
      q(in1) match {
        case Success(y, in2) => Success(compose(x, y), in2)
        case f @ Failure(_, _) => if (retry) f else f.noRetry
      }
    case f @ Failure(_, _) => f
  }

  /** A helper method for functional composition of (unit-)parsers */
  private def mapfun[T, U](p: Input => ParseResult[T])
                          (f: T => U)
                          (in: Input): ParseResult[U] = p(in) match {
    case s1 @ Success(x, in1) => Success(f(x), in1)
    case f @ Failure(_, _) => f
  }

  /** A parser matching input elements which satisfy a given predicate
   *  @param  kind   The element kind, used for error messages
   *  @param  p      A predicate which determins which elements match.
   *  
   *  elem(kind, e) succeeds if the input starts with an element `e' for which p(e)
   *  is true. 
   */
  def elem(kind: String, p: Elem => boolean) = new Parser[Elem] {
    def apply(in: Input) =
      if (p(in.first)) Success(in.first, in.rest)
      else Failure(kind+" expected", in)
  }

  /** The parser that matches only given element `e' 
   *  The method is implicit to that elements can be automatically
   *  lifted to their unit-parsers.
   */
  implicit def accept(e: Elem): UnitParser = new UnitParser {
    def apply(in: Input) =
      if (in.first == e) Success((), in.rest)
      else Failure("`"+e+"' expected but " + in.first + " found", in)
  }

  /** A parser which always fails 
   *  @param msg   The error message describing the failure. */
  def failure(msg: String) = new Parser[Nothing] {
    def apply(in: Input) = Failure(msg, in)
  }

  /** A unit-parser which always fails 
   *  @param msg   The error message describing the failure. */
  def fail(msg: String) = new UnitParser {
    def apply(in: Input) = Failure(msg, in)
  }

  /** A parser which always succeeds 
   *  @param out   The output value of the parser. */
  def success[T](v: T) = new Parser[T] {
    def apply(in: Input) = Success(v, in)
  }

  /** A parser generator for repetitions.
   *  
   *  rep(p)   runs `p' zero or more times on input until no more match
   */
  def rep[T](p: Parser[T]): Parser[List[T]] = rep1(p) | success(List())

  /** A parser generator for non-empty repetitions.
   *  
   *  rep1(p)   runs `p' one or more times on input until no more match
   */
  def rep1[T](p: Parser[T]): Parser[List[T]] = 
    p ++ rep(p) ^^ { case Pair(x, xs) => x :: xs }

  /** A parser generator for optional sub-phrases.
   *  
   *  opt(p)   runs `p' zero or once on input
   */
  def opt[T](p: Parser[T]): Parser[Option[T]] = 
    p ^^ (x => Some(x)) | success(None)


  /** A parser that decorates its result with the start position
   *  of the input consumed by the parser `p'. The result of `p' must conform to 
   *  type Positional.
   *  If the result value carries already a position, it is kept unchanged.
   */
  def positioned[T <: Positional](p: Parser[T]): Parser[T] = new Parser[T] {
    def apply(in: Input) = p(in) match {
      case Success(t, in1) => Success(if (t.pos == NoPosition) t setPos in.pos else t, in1)
      case f @ Failure(_, _) => f
    }
  }

  /** A parser that succeeds iff its unit-parser argument `p' succeeds.
   *  In that case, the start position of the input consumed by `p' is returned as result
   */
  def positioned(p: UnitParser) = new Parser[Position] {
    def apply(in: Input) = p(in) match {
      case Success(_, in1) => Success(in.pos, in1) 
      case f @ Failure(_, _) => f
    }
  }
  
  trait OnceParser[+T] extends Parser[T] {
    override def ++ [U](q: => Parser[U]): Parser[Pair[T, U]] = new Parser[Pair[T, U]] with OnceParser[Pair[T, U]] {
        def apply(in: Input) = seq(OnceParser.this, q)((x, y) => Pair(x, y))(false)(in)
    }
    override def ++ [Q <% UnitParser](q: Q): Parser[T] = new Parser[T] with OnceParser[T] {
      def apply(in: Input) = seq(OnceParser.this, q)((x, y) => x)(false)(in)
    }
  }
  
  trait UnitOnceParser extends UnitParser {
    override def ++ [U](q: => Parser[U]): Parser[U] = new Parser[U] with OnceParser[U]{
      def apply(in: Input): ParseResult[U] = seq(UnitOnceParser.this, q)((x, y) => y)(false)(in)
    }
    override def ++ [A <% UnitParser](q: A): UnitParser = new UnitOnceParser {
      def apply(in: Input): ParseResult[unit] = seq(UnitOnceParser.this, q)((x, y) => y)(false)(in)
    }
  }
 }
  
